import hashlib
import html
import re
import time

import aiohttp

from . import salt

session_id_regex = r'"web-authorize-session-id":"([\w]+=:\d+:[\w\d]+)"'


async def get_session_id(url: str) -> str:
    async with aiohttp.ClientSession() as session:
        res = await session.get(url)
        match = re.search(session_id_regex, html.unescape(await res.text()))
        if match is None:
            raise Exception("Session ID not found")
        return match.group(1)


async def generate_header(path: str, session_id: str, method="POST") -> dict[str, str]:
    async with aiohttp.ClientSession() as session:
        res = await session.get(
            f"https://twitcasting.tv/js/v1/PlayerPage2.js?{int(time.time())}"
        )
        seed = salt.get_salt(await res.text())
        timestamp = int(time.time())
        text = f"{seed}{timestamp}{method}{path}{session_id}"
        h = hashlib.sha256()
        h.update(text.encode())
        hashed_text = h.hexdigest()
        authorizekey = f"{timestamp}.{hashed_text}"

        return {"x-web-authorizekey": authorizekey, "x-web-sessionid": session_id}


async def get_token(movie_id: int, session_id: str) -> str:
    async with aiohttp.ClientSession() as session:
        path = f"/movies/{movie_id}/token"

        res = await session.post(
            f"https://frontendapi.twitcasting.tv{path}",
            headers={
                "Content-Type": "multipart/form-data",
            }
            | await generate_header(path, session_id),
        )
        if res.status // 100 != 2:
            raise Exception("Failed to get token")
        data = await res.json()
        return data["token"]



from typing import TypedDict


class Movie(TypedDict):
    id: int
    live: bool


class Hls(TypedDict):
    host: str
    proto: str
    source: bool


class Fmp4(TypedDict):
    host: str
    proto: str
    source: bool
    mobilesource: bool


class Streams(TypedDict):
    main: str
    base: str


class Llfmp4(TypedDict):
    streams: Streams


class StreamServer(TypedDict):
    movie: Movie
    hls: Hls
    fmp4: Fmp4
    llfmp4: Llfmp4


class EventPubSubUrl(TypedDict):
    url: str



from omu_chat.model import Provider
from omu_chatprovider.chatprovider import BASE_PROVIDER_IDENTIFIER
from omu_chatprovider.helper import HTTP_REGEX

from .version import VERSION

PROVIDER_ID = BASE_PROVIDER_IDENTIFIER / "twitcasting"
PROVIDER = Provider(
    id=PROVIDER_ID,
    url="https://twitcasting.tv/",
    name="TwitCasting",
    version=VERSION,
    repository_url="https://github.com/OMUAPPS/omuapps-python/tree/master/packages-py/chat-twitcasting",
    regex=HTTP_REGEX + r"twitcasting\.tv/(?P<user_id>[^/]+)",
)
BASE_HEADERS = {"User-Agent": f"OMUAPPS/{VERSION} twitcasting provider"}



from omu import Omu
from omu_chat import Chat
from omu_chat.model import Channel, Provider, Room
from omu_chatprovider.helper import get_session
from omu_chatprovider.service import FetchedRoom, ProviderService

from .api import TwitcastingAPI
from .chat import TwitcastingChat
from .const import (
    PROVIDER,
)


class TwitcastingChatService(ProviderService):
    def __init__(self, omu: Omu, chat: Chat):
        self.omu = omu
        self.chat = chat
        self.session = get_session(omu, PROVIDER)
        self.api = TwitcastingAPI(omu, self.session)

    @property
    def provider(self) -> Provider:
        return PROVIDER

    async def fetch_rooms(self, channel: Channel) -> list[FetchedRoom]:
        room = await self.api.fetch_room(channel)

        async def create(room=room) -> TwitcastingChat:
            return TwitcastingChat(
                self.session,
                self.chat,
                room,
                channel,
            )

        return [
            FetchedRoom(
                room,
                create,
            )
        ]

    async def is_online(self, room: Room) -> bool:
        return True



import time
from datetime import datetime

import aiohttp
from omu import Omu
from omu_chat.model.channel import Channel
from omu_chat.model.room import Room

from .const import PROVIDER
from .token import get_session_id, get_token
from .types import movieinfo, types


class TwitcastingAPI:
    def __init__(self, omu: Omu, session: aiohttp.ClientSession):
        self.omu = omu
        self.session = session

    async def fetch_room(self, channel: Channel) -> Room:
        user_id = channel.id.path[-1]
        res = await self.session.get(
            "https://twitcasting.tv/streamserver.php",
            params={
                "target": user_id,
                "mode": "client",
                "player": "pc_web",
            },
        )
        stream_server: types.StreamServer = await res.json()
        movie = stream_server["movie"]
        token = await get_token(
            movie["id"],
            await get_session_id(f"https://twitcasting.tv/{user_id}"),
        )
        response = await self.session.get(
            f"https://frontendapi.twitcasting.tv/movies/{movie['id']}/status/viewer",
            params={
                "token": token,
                "__n": int(time.time() * 1000),
            },
        )
        movie_info: movieinfo.MovieInfo = await response.json()
        movie = movie_info["movie"]
        room = Room(
            id=channel.id / str(movie["id"]),
            channel_id=channel.id,
            provider_id=PROVIDER.id,
            connected=False,
            status="offline",
            created_at=datetime.now(),
            metadata={
                "title": movie.get("title", None),
                "description": movie.get("telop", None),
                "thumbnail": f"https://twitcasting.tv/userajax.php?c=updateindexthumbnail&m={movie["id"]}&u={user_id}",
                "url": f"https://twitcasting.tv/{user_id}/movie/{movie["id"]}",
                "viewers": movie["viewers"]["current"],
            },
        )
        return room



from typing import Literal, TypedDict


class Sender(TypedDict):
    """
    {
        "id": "namae_id",
        "name": "なまえ名前Name",
        "screenName": "namae_id",
        "profileImage": "https://imagegw02.twitcasting.tv/image3s/pbs.twimg.com/profile_images/([0-9]{19})/3f5ZZuhS_normal.jpg",
        "grade": 0,
        "m": true
    }
    """

    id: str
    name: str
    screenName: str
    profileImage: str
    grade: int


class CommentEvent(TypedDict):
    """
    {
        "type": "comment",
        "id": 12345678901,
        "message": "メッセージMessage",
        "createdAt": 1703125960000,
        "author": Sender,
        "numComments": 78463
    }
    """

    type: Literal["comment"]
    id: int
    message: str
    createdAt: int
    author: Sender
    numComments: int


class Item(TypedDict):
    """
    {
        "name": "お茶",
        "image": "https://s01.twitcasting.tv/img/item_tea.png",
        "detailImage": "",
        "effectCommand": "",
        "showsSenderInfo": true
    }
    """

    name: str
    image: str
    detailImage: str
    effectCommand: str
    showsSenderInfo: bool


class ScoreItem(TypedDict):
    """
    {
        "title": "🍡",
        "text": "2"
    }
    """

    title: str
    text: str


class GiftEvent(TypedDict):
    """
    {
        "id": "012abcdef.0123456789abcdef.12345678",  # [a-f0-9]{8}.[a-f0-9]{16}.[0-9]{8}
        "type": "gift",
        "message": "(+🍡2)",
        "plainMessage": "",
        "item": Item,
        "sender": Sender,
        "createdAt": 1703126345000,
        "score_items": [ScoreItem]
    }
    """

    type: Literal["gift"]
    id: str
    message: str
    plainMessage: str
    item: Item
    sender: Sender
    isForMovie: bool
    isPaidGift: bool
    createdAt: int
    score_items: list[ScoreItem]


Event = CommentEvent | GiftEvent

"""
[Event...]
"""
type EventJson = list[Event]



from typing import TypedDict


class Category(TypedDict):
    id: str
    name: str


class Viewers(TypedDict):
    current: int
    total: int


class Movie(TypedDict):
    id: int
    title: str
    telop: str
    category: Category
    viewers: Viewers
    hashtag: str
    pin_message: str


class MovieInfo(TypedDict):
    update_interval_sec: int
    movie: Movie



"""
Modified version of https://github.com/HoloArchivists/TwcLazer/blob/main/utils/Salt.py
"""

import base64
import re


def decode_string(text):
    """
    Base64 decode with case swap for extra obfuscation
    """
    text = text + "=" * (len(text) % 4)
    return base64.b64decode(bytes(text, "ascii").swapcase()).decode("ascii")


def get_encoded_array(js):
    # array of encoded strings with function names and constants, used
    # in various places of the script. One of them is the salt.
    # Surprizingly, "const e" doesn't change over the versions of the script
    re_constants = r"const e=\[((\"[a-zA-Z0-9]+\",)+\"[a-zA-Z0-9]+\")\]"
    match = re.search(re_constants, js)
    if match is None:
        raise Exception("Salt extraction failed: constants not found")
    return [x.strip('"') for x in match.groups()[0].split(",")]


def get_base_offset(js):
    # strings on the array are accessed through specifying its index,
    # and function wrapping the array substracts constant value from
    # index it receives, which is different every script version
    re_offset = r"const [a-z]=[a-z]\(\);return [a-z]=function\([a-z],[a-z]\){let [a-z]=[a-z]\[[a-z]-=(\d+)\];"
    offset = re.findall(re_offset, js)[0]
    return int(offset)


def get_shift_offsets(js):
    # picking up indexes from the
    # "... parseInt(n(429)) / 1 + parseInt(n(448)) / 2 * ( - parseInt(n(428)) / 3) ..."
    re_offsets_line = r"if\(\d+==(.+)break;[a-z].push\([a-z].shift\(\)\)}"
    match = re.search(re_offsets_line, js)
    if match is None:
        raise Exception("Salt extraction failed: offsets not found")
    offsets_line = match.groups()[0]
    re_offsets = r"parseInt\([a-z]\((\d+)\)\)"
    offsets = re.findall(re_offsets, offsets_line)
    return [int(offset) for offset in offsets]


def starts_with_number(text):
    # cutting corners with parseInt() implementation: the way it works
    # it will throw away any non-numeric characters at the end of the string,
    # and for this case only the fact that it returns a number matters, not the number itself
    return re.match(r"\d+", text) is not None


def decode_array(js):
    # the array of encoded strings is also additionally rotated (circularly shifted), so order of
    # strings in runtime differs from that in code. The rule to determine correct
    # number of rotation steps is to try every possible position and pick the one,
    # where all (decoded) strings in array that are located at offsets used in the
    # "... parseInt(n(429)) / 1 + parseInt(n(448)) / 2 * ( - parseInt(n(428)) / 3) ..."
    # expression are successfully parseInt'ed as a number
    candidates = []
    base_offset = get_base_offset(js)
    test_points_offsets = [offset - base_offset for offset in get_shift_offsets(js)]
    array = [decode_string(x) for x in get_encoded_array(js)]
    for _ in range(len(array)):
        strings_on_offsets = [array[i] for i in test_points_offsets]
        if all(starts_with_number(line) for line in strings_on_offsets):
            candidates.append(array.copy())
        array.append(array.pop(0))
    if len(candidates) == 0:
        raise Exception(
            "Salt extraction failed: none of possible array positions matched"
        )
    elif len(candidates) == 1:
        return candidates[0]
    else:
        print(
            f"Salt extraction: found {len(candidates)} possible offsets, using first one"
        )
        return candidates[0]


def get_salt_offset(js):
    # extracting array offset for salt value from the place in code that makes use of it.
    # Since array can be decoded and rotated to right position, all that is left is to
    # get the salt value out of it by its offset
    re_salt_offset = r'null!==\([a-z]=[a-z]\((\d+)\)\)\?[a-z]:"",[a-z]\)'
    match = re.search(re_salt_offset, js)
    if match is None:
        raise Exception("Salt extraction failed: salt offset not found")
    return int(match.groups()[0])


def get_salt(js):
    salt_offset = get_salt_offset(js) - get_base_offset(js)
    strings = decode_array(js)
    salt = strings[salt_offset]
    return salt



from __future__ import annotations

import asyncio
import random
import string
import time
from datetime import datetime

import aiohttp
from aiohttp import WSMsgType
from loguru import logger
from omu.identifier import Identifier
from omu_chat.chat import Chat
from omu_chat.model import Author, Channel, Message, Room
from omu_chat.model.content import Root, Text
from omu_chat.model.gift import Gift
from omu_chatprovider.service import ChatService
from omu_chatprovider.tasks import Tasks
from yarl import URL

from .const import PROVIDER
from .types import events, types


class TwitcastingChat(ChatService):
    def __init__(
        self,
        session: aiohttp.ClientSession,
        chat: Chat,
        room: Room,
        channel: Channel,
    ):
        self.session = session
        self.movie_id = room.id.path[-1]
        self.chat = chat
        self._room = room
        self.channel = channel
        self._closed = False
        self.tasks = Tasks(asyncio.get_event_loop())

    @property
    def room(self) -> Room:
        return self._room

    @property
    def closed(self) -> bool:
        return self._closed

    async def run(self):
        while True:
            socket = await self.create_socket()
            self.room.status = "online"
            await self.chat.rooms.update(self.room)
            await self.listen(socket)
            self.room.status = "offline"
            await self.chat.rooms.update(self.room)

    async def create_socket(self):
        socket = await self.session.ws_connect(
            await self.fetch_ws(),
            params={
                "gift": "1",
            },
        )
        return socket

    async def listen(self, socket: aiohttp.ClientWebSocketResponse):
        while True:
            message = await socket.receive()
            if message.type == WSMsgType.CLOSED:
                break
            if message.type == WSMsgType.ERROR:
                break
            if message.type == WSMsgType.CLOSE:
                break
            data = message.json()
            authors: list[Author] = []
            messages: list[Message] = []
            for event in self.validate_event(data):
                await self.process_message(event, authors, messages)
            if len(authors) > 0:
                new_authors = [
                    author
                    for author in authors
                    if await self.chat.authors.get(author.id.key()) is None
                ]
                await self.chat.authors.add(*new_authors)
            if len(messages) > 0:
                await self.chat.messages.add(*messages)

    def validate_event(self, data: events.EventJson) -> list[events.Event]:
        if not isinstance(data, list):
            logger.warning(f"Invalid event: {data}")
            return []
        return data

    async def process_message(
        self,
        event: events.Event,
        authors: list[Author],
        messages: list[Message],
    ):
        if event["type"] == "comment":
            author = await self._parse_sender(event["author"])
            created_at = self._parse_created_at(event)
            message = Message(
                id=self.room.id / str(event["id"]),
                room_id=self.room.id,
                content=Root([Text.of(event["message"])]),
                author_id=author.id,
                created_at=created_at,
            )
            authors.append(author)
            messages.append(message)
        elif event["type"] == "gift":
            author = await self._parse_sender(event["sender"])
            created_at = self._parse_created_at(event)
            gifts = [
                self._parse_item(event["item"]),
                *self._parse_items(event.get("score_items", [])),
            ]
            message = Message(
                id=self.room.id / str(event["id"]),
                room_id=self.room.id,
                content=Root([Text.of(event["message"])]),
                author_id=author.id,
                created_at=created_at,
                gifts=gifts,
            )
            authors.append(author)
            messages.append(message)
        else:
            raise Exception(f"Unknown event type: {event['type']}")

    def _parse_item(self, item: events.Item) -> Gift:
        return Gift(
            id=item["name"],
            name=item["name"],
            amount=1,
            image_url=self.get_full_url(item["image"]),
            is_paid=False,
        )

    def get_full_url(self, path: str) -> str:
        url = URL(path)
        return str(
            URL.build(
                scheme=url.scheme or "https",
                host=url.host or "s01.twitcasting.tv",
                path=url.path,
            )
        )

    def _parse_items(self, items: list[events.ScoreItem]) -> list[Gift]:
        gifts = []
        for item in items:
            amount = 1
            if item["text"] and item["text"].isdigit():
                amount = int(item["text"])
            gifts.append(
                Gift(
                    id=item["title"],
                    name=item["title"],
                    amount=amount,
                    is_paid=False,
                )
            )
        return gifts

    async def _parse_sender(self, sender: events.Sender) -> Author:
        author = Author(
            id=self.get_sender_id(self.channel.id, sender["id"]),
            provider_id=PROVIDER.id,
            name=sender["name"],
            avatar_url=sender["profileImage"],
            metadata={
                "url": f"https://twitcasting.tv/{sender['id']}",
            },
        )
        return author

    def get_sender_id(self, channel_id: Identifier, sender_id: str) -> Identifier:
        """
        c:channel_id -> c/channel_id
        channel_id -> _/channel_id
        """
        if ":" not in sender_id:
            return channel_id / "_" / sender_id
        type, id = sender_id.split(":", 1)
        if ":" in id:
            raise Exception(f"Invalid sender id: {sender_id}")
        return channel_id / type / id

    def _parse_created_at(self, event: events.Event) -> datetime:
        created_at = datetime.fromtimestamp(event["createdAt"] / 1000)
        return created_at

    async def fetch_ws(self) -> str:
        key = "WebKitFormBoundary" + "".join(
            random.choices(string.ascii_letters + string.digits, k=16)
        )
        n = int(time.time() * 1000)
        form_data = f"""------{key}
Content-Disposition: form-data; name="movie_id"

{self.movie_id}
------{key}
Content-Disposition: form-data; name="__n"

{n}
------{key}
Content-Disposition: form-data; name="password"


------{key}--"""
        res = await self.session.post(
            "https://twitcasting.tv/eventpubsuburl.php",
            data=form_data,
            headers={
                "Content-Type": f"multipart/form-data; boundary=----{key}",
            },
        )
        data: types.EventPubSubUrl = await res.json()
        return data["url"]

    async def stop(self):
        self.tasks.terminate()
        self.room.status = "offline"
        await self.chat.rooms.update(self.room)
